"""EYWA Reacher client"""

__author__ = "Robert Gersak"
__email__ = "r.gersak@gmail.com"
__license__ = "MIT"
__status__ = "Development"
__version__ = "0.0.8"


import time
import datetime
import json
import sys
import threading
import traceback


def check_id(id, allow_empty=False):
    if (id is not None or not allow_empty) and not isinstance(id, (int, str)):
        raise TypeError("id must be an integer or string, got {} ({})".format(id, type(id)))

def check_method(method):
    if not isinstance(method, str):
        raise TypeError("method must be a string, got {} ({})".format(method, type(method)))

def check_code(code):
        if not isinstance(code, int):
            raise TypeError("code must be an integer, got {} ({})".format(id, type(id)))

        if not get_error(code):
            raise ValueError("unknown code, got {} ({})".format(code, type(code)))

def generate_request(method, id=None, params=None):
        try:
            check_method(method)
            check_id(id, allow_empty=True)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        req = "{{\"jsonrpc\":\"2.0\",\"method\":\"{}\"".format(method)

        if id is not None:
            if isinstance(id, str):
                id = json.dumps(id)
            req += ",\"id\":{}".format(id)

        if params is not None:
            try:
                req += ",\"params\":{}".format(json.dumps(params))
            except Exception as e:
                raise RPCParseError(str(e))

        req += "}"

        return req


def generate_response(id, result):
        try:
            check_id(id)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        # encode string ids
        if isinstance(id, str):
            id = json.dumps(id)

        # build the response string
        try:
            res = "{{\"jsonrpc\":\"2.0\",\"id\":{},\"result\":{}}}".format(id, json.dumps(result))
        except Exception as e:
            raise RPCParseError(str(e))

        return res

def generate_error(id, code, data=None):
        try:
            check_id(id)
            check_code(code)
        except Exception as e:
            raise RPCInvalidRequest(str(e))

        # build the inner error data
        if (get_error(code) is not None):
            message = get_error(code).title
        else:
            message = 'Unknown error code'
        err_data = "{{\"code\":{},\"message\":\"{}\"".format(code, message)

        # insert data when given
        if data is not None:
            try:
                err_data += ",\"data\":{}}}".format(json.dumps(data))
            except Exception as e:
                raise RPCParseError(str(e))
        else:
            err_data += "}"

        # encode string ids
        if isinstance(id, str):
            id = json.dumps(id)

        # start building the error string
        err = "{{\"jsonrpc\":\"2.0\",\"id\":{},\"error\":{}}}".format(id, err_data)

        return err


EMPTY_RESULT = object()


connection_thread=None

class EYWA():

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(EYWA, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        # other attributes
        self._i = -1
        self._callbacks = {}
        self._handlers= {}
        self._results = {}

    def __del__(self):
        if connection_thread:
            connection_thread.stop()

    @classmethod
    def request(cls, method, params, timeout = None):
        # check if the call is a notification
        eywa = EYWA()
        # create a new id for requests expecting a response
        eywa._i += 1
        id = eywa._i

        now = time.time();

        # store an empty result for the meantime
        eywa._results[id] = EMPTY_RESULT

        # create the request
        req = generate_request(method, id=id, params=params)
        eywa._write(req)

        result  = EMPTY_RESULT
        # blocking return value behavior
        while True:
            if eywa._results[id] != EMPTY_RESULT:
                result = eywa._results[id]
                del eywa._results[id]
                if isinstance(result, Exception):
                   raise result
                else:
                    return result
            elif timeout is not None  and time.time() - now > timeout:
                raise ValueError('EYWA silent error. Timeout occured for:\n' + req)
            else:
                time.sleep(0.1)

    @classmethod
    def notify(cls, method, data = None):
        eywa = EYWA()
        req = generate_request(method, None, data)
        eywa._write(req)

    def _handle(self, line):
        """
        Handles an incoming *line* and dispatches the parsed object to the request, response, or
        error handlers.
        """
        obj = None
        try:
            obj = json.loads(line)
        except Exception as e:
            # print('Line:\n' + line, file=sys.stderr)
            traceback.print_exception(e)
            pass

        # dispatch to the correct handler
        if obj is None:
            pass
        elif "method" in obj:
            # request
            self._handle_request(obj)
        elif "error" not in obj:
            # response
            self._handle_response(obj)
        else:
            # error
            self._handle_error(obj)

    def _handle_request(self, req):
        try:
            method = self._route(req["method"])
            result = method(req["params"])
            if "id" in req:
                res = generate_response(req["id"], result)
                self._write(res)
        except Exception as e:
            if "id" in req:
                if isinstance(e, RPCError):
                    err = generate_error(req["id"], e.code, e.data)
                else:
                    err = generate_error(req["id"], -32603, str(e))
                self._write(err)

    def _handle_response(self, res):
        if res["id"] in self._results:
            self._results[res["id"]] = res["result"]

        if res["id"] in self._callbacks:
            callback = self._callbacks[res["id"]]
            del self._callbacks[res["id"]]
            callback(None, res["result"])

    def _handle_error(self, res):
        err = res["error"]
        error = get_error(err["code"])(err.get("data", err["message"]))

        # set the error
        if res["id"] in self._results:
            self._results[res["id"]] = error

        # lookup and invoke the callback
        if res["id"] in self._callbacks:
            callback = self._callbacks[res["id"]]
            del self._callbacks[res["id"]]
            callback(error, None)

    def _route(self, method):
        if method in self._handlers:
            handler = self._handlers[method]
            return handler 
        raise RPCMethodNotFound(data=method)

    def _write(self, s):
        """
        Writes a string *s* to the output stream.
        """
        sys.stdout.write(s + "\n")
        sys.stdout.flush()


class Line(threading.Thread):
    def __init__(self, tap, name="EYWAonline", interval=0.1, start=True):

        super(Line, self).__init__()

        # store attributes
        self.tap = tap
        self.name = name
        self.interval = interval
        self.daemon = True

        # register a stop event
        self._stop = threading.Event()

        if start:
            self.start()

    def start(self):
        super(Line, self).start()

    def stop(self):
        self._stop.set()

    def run(self):
        self._stop.clear()
        # print('starting read loop')
        while not self._stop.is_set():
            line= None
            try:
                # print('Reading sys.stdin.line')
                line = sys.stdin.readline()
                # print('Line length %d', len(line))
            except IOError:
                # prevent residual race conditions occurring when stdin is closed externally
                pass
            if line:
                # line = line.strip()
                self.tap._handle(line)
            else:
                self._stop.wait(self.interval)


class RPCError(Exception):

    """
    Base class for RPC errors.

    .. py:attribute:: message

       The message of this error, i.e., ``"<title> (<code>)[, data: <data>]"``.

    .. py:attribute:: data

       Additional data of this error. Setting the data attribute will also change the message
       attribute.
    """

    def __init__(self, data=None):
        # build the error message
        message = "{} ({})".format(self.title, self.code)
        if data is not None:
            message += ", data: {}".format(data)
        self.message = message

        super(RPCError, self).__init__(message)

        self.data = data

    def __str__(self):
        return self.message


error_map_distinct = {}
error_map_range = {}


def is_range(code):
    return (
        isinstance(code, tuple) and
        len(code) == 2 and
        all(isinstance(i, int) for i in code) and
        code[0] < code[1]
    )


def register_error(cls):
    """
    Decorator that registers a new RPC error derived from :py:class:`RPCError`. The purpose of
    error registration is to have a mapping of error codes/code ranges to error classes for faster
    lookups during error creation.

    .. code-block:: python

       @register_error
       class MyCustomRPCError(RPCError):
           code = ...
           title = "My custom error"
    """
    # it would be much cleaner to add a meta class to RPCError as a registry for codes
    # but in CPython 2 exceptions aren't types, so simply provide a registry mechanism here
    if not issubclass(cls, RPCError):
        raise TypeError("'{}' is not a subclass of RPCError".format(cls))

    code = cls.code

    if isinstance(code, int):
        error_map = error_map_distinct
    elif is_range(code):
        error_map = error_map_range
    else:
        raise TypeError("invalid RPC error code {}".format(code))

    if code in error_map:
        raise AttributeError("duplicate RPC error code {}".format(code))

    error_map[code] = cls

    return cls


def get_error(code):
    """
    Returns the RPC error class that was previously registered to *code*. *None* is returned when no
    class could be found.
    """
    if code in error_map_distinct:
        return error_map_distinct[code]

    for (lower, upper), cls in error_map_range.items():
        if lower <= code <= upper:
            return cls

    return None


@register_error
class RPCParseError(RPCError):

    code = -32700
    title = "Parse error"


@register_error
class RPCInvalidRequest(RPCError):

    code = -32600
    title = "Invalid Request"


@register_error
class RPCMethodNotFound(RPCError):

    code = -32601
    title = "Method not found"


@register_error
class RPCInvalidParams(RPCError):

    code = -32602
    title = "Invalid params"


@register_error
class RPCInternalError(RPCError):

    code = -32603
    title = "Internal error"


@register_error
class RPCServerError(RPCError):

    code = (-32099, -32000)
    title = "Server error"





# EYWA INTERNALS

class Sheet ():
    def __init__(self, name = 'Sheet'):
        self.name = name
        self.rows = []
        self.columns = []
    def add_row(self,row):
        self.rows.append(row)
    def remove_row(self,row):
        self.rows.remove(row)
    def set_columns(self, columns):
        self.columns = columns
    def toJSON(self):
        return json.dumps(self, default=lambda o:o.__dict__)


class Table ():
    def __init__(self, name = 'Table'):
        self.name = name
        self.sheets= []
    def add_sheet(self,sheet):
        self.sheets.append(sheet)
    def remove_sheet(self,idx=0):
        self.sheets.pop(idx)
    def toJSON(self):
        return json.dumps(self, default=lambda o:o.__dict__)


# TODO finish task reporting
class TaskReport():
    def __init__(self,message, data=None, image=None):
        self.message = message
        self.data = data
        self.image = image

# ws1 = Sheet('miroslav')
# ws1.add_row({'slaven':1,'belupo':2})
# ws1.add_row({'slaven':30,'belupo':0})


# t1 = Table('TEST')
# t1.add_sheet(ws1)

# print(t1.toJSON())
# print(json.dumps({'a':2,'b':'4444'}))


SUCCESS = "SUCCESS"
ERROR = "ERROR"
PROCESSING = "PROCESSING"
EXCEPTION = "EXCEPTION"

def log(event="INFO",
        message=None,
        data=None,
        duration=None,
        coordinates=None,
        time=None):

    if (time == None):
        time= datetime.datetime.utcnow().isoformat()

    if event not in ["INFO", "ERROR", "WARN", "DEBUG", "TRACE"]:
        raise ValueError(f"The event '{event}' is not one of allowed event types [INFO, ERROR, WARN, DEBUG, TRACE].")
    EYWA.notify("task.log", {"time": time, "event":event,"message":message,
        "data":data,"coordinates":coordinates,"duration":duration})

def info(message,data=None):
    log("INFO", message, data)

def error(message,data=None):
    log("ERROR", message, data)

def warn(message,data=None):
    log("WARN",message,data)

def debug(message,data=None):
    log("DEBUG",message,data)

def trace(message,data=None):
    log("TRACE",message,data)

def report(message,data=None,image=None):
    EYWA.notify("task.report",
                {"message":message,
                 "data": data,
                 "image":image})

def close(status=SUCCESS):
    EYWA.notify("task.close", {"status":status})
    if (status == ERROR):
        exit_status=1
    else:
        exit_status=0
    sys.exit(exit_status)


def get_task():
    return EYWA.request("task.get", {})


def update_task(status=PROCESSING):
    if status not in ["SUCCESS", "ERROR", "PROCESSING", "EXCEPTION"]:
        raise ValueError(f"The status {status} is not one of allowed status types [SUCCESS, ERROR, PROCESSING, EXCEPTION]")
    EYWA.notify("task.update",{"status":status})


def return_task():
    EYWA.notify("task.return")
    sys.exit(0)

def graphql(query, timeout = None):
    return EYWA.request("eywa.datasets.graphql", query, timeout)

eywa = EYWA()
connection_thread=Line(eywa)



import eywa
from datetime import datetime


query = """
mutation($example:TaskInput!) {
  syncTask(task:$example) {
    euuid
  }
}
"""


response = eywa.graphql({'query': query, 'variables': {
    "example": {
        "euuid": "ff78873b-15dc-43e1-b845-93064bdeccc1",
        "message": "Testing Python reacher client",
        "data": {"a": 100,
                 "drvo": "hrast",
                 "kamen": "bacim"},
        "started": datetime(2000, 2, 3, 4, 5, 6).isoformat()
        }
    }}, 2)


print('Response:\n' + str(response))

eywa.close()



import eywa
from selenium import webdriver

eywa.error("Some shit happened")
eywa.info("Normal as usual")
eywa.warn("Uuuu you should be scared")
eywa.update_task()
# task.close(task.ERROR)

print('Evo nekog texta')
eywa.update_task(status=eywa.PROCESSING)
eywa.info("Opening Chrome browser")
browser = webdriver.Chrome()
eywa.info("Chrome opened")
eywa.info("Navigation to www.google.com")
browser.get("http://www.google.com")
eywa.info("Google visible")
browser.close()
eywa.info("Browser closed")
eywa.report("Everything went just fine",{'hanky':'dory'})
# eywa.close(eywa.SUCCESS)



from faker import Faker
import eywa


# Create a Faker instance and set the locale to 'en_US'
fake = Faker('en_US')

# Generate and print a random name
print("Name:", fake.name())

# Generate and print a random address
print("Address:", fake.address())

# Generate and print a random email
print("Email:", fake.email())

# Generate and print a random date
print("Birthdate:", fake.date_of_birth())

# You can also create multiple fake data entries in a loop
for _ in range(5):
    print("Random Name:", fake.name())
    print("Random Email:", fake.email())
    print("Random Address:", fake.address())
    print("Random Birthdate:", fake.date_of_birth())
    print("-" * 20)



import sys
import json

query = """
{
    searchUser {
        euuid
        name
        type
        modified_on
        modified_by {
            name
        }
    }
}"""

sys.stdout.write(str(json.dumps({
    'jsonrpc': "2.0",
    'id':0,
    'method': 'eywa.datasets.graphql',
    'params': {
        'query': query,
        'variables': {
            'a': 10,
            'b':20
            }
        }})) +
    "\n")
sys.stdout.flush()


response = sys.stdin.readline()

sys.stdout.write('EYWA')
sys.stdout.write(response)
sys.stdout.flush()


# {"jsonrpc":"2.0","id":0,"result":100} 
# {"jsonrpc":"2.0","id":0,"error": {"code": -32602, "message": "Fucker"}} 



import asyncio
import sys
import json
from nanoid import generate as nanoid

rpc_callbacks = {}
handlers = {}


def handle_data(data):
    method = data.get("method")
    id_ = data.get("id")
    result = data.get("result")
    error = data.get("error")
    if method:
        handle_request(data)
    elif result and id_:
        handle_response(data)
    elif error and id_:
        handle_response(data)
    else:
        print('Received invalid JSON-RPC:\n', data)


def handle_request(data):
    method = data.get("method")
    handler = handlers.get(method)
    if handler:
        handler(data)
    else:
        print(f"Method {method} doesn't have registered handler")


def handle_response(data):
    id_ = data.get("id")
    callback = rpc_callbacks.get(id_)
    if callback is not None:
        callback.set_result(data)
        # print(f'Handling response for {callback}')
    else:
        print(f'RPC callback not registered for request with id = {id_}')


async def send_request(data):
    id_ = nanoid()
    id_ = 10
    data["jsonrpc"] = "2.0"
    data["id"] = id_
    future = asyncio.Future()
    rpc_callbacks[id_] = future
    sys.stdout.write(json.dumps(data) + "\n")
    result = await future
    del rpc_callbacks[id_]
    return result


def send_notification(data):
    data["jsonrpc"] = "2.0"
    sys.stdout.write(json.dumps(data) + "\n")


def register_handler(method, func):
    handlers[method] = func


class LargeBufferStreamReader(asyncio.StreamReader):
    # Default limit set to 1 MB here.
    def __init__(self, limit=1024*1024*10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._limit = limit


async def read_stdin():
    reader = LargeBufferStreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)
    await asyncio.get_event_loop().connect_read_pipe(lambda: protocol, sys.stdin)

    while True:
        try:
            raw_json = await asyncio.wait_for(reader.readline(), timeout=2)
            json_data = json.loads(raw_json.decode().strip())
            handle_data(json_data)
            await asyncio.sleep(0.5)
        except asyncio.TimeoutError:
            await asyncio.sleep(0.5)


# Additional functions

def log(event="INFO", message="", data=None, duration=None, coordinates=None, time=None):
    if time is None:
        from datetime import datetime
        time = datetime.now()

    send_notification({
        "method": "task.log",
        "params": {
            "time": time,
            "event": event,
            "message": message,
            "data": data,
            "coordinates": coordinates,
            "duration": duration
        }
    })


def info(message, data=None):
    log(event="INFO", message=message, data=data)


def error(message, data=None):
    log(event="ERROR", message=message, data=data)


def warn(message, data=None):
    log(event="WARN", message=message, data=data)


def debug(message, data=None):
    log(event="DEBUG", message=message, data=data)


def trace(message, data=None):
    log(event="TRACE", message=message, data=data)


def report(message, data=None, image=None):
    send_notification({
        'method': 'task.report',
        'params': {
            'message': message,
            'data': data,
            'image': image
        }
    })


def close_task(status="SUCCESS"):
    send_notification({
        'method': 'task.close',
        'params': {
            'status': status
        }
    })

    if status == "SUCCESS":
        sys.exit(0)
    else:
        sys.exit(1)


def update_task(status="PROCESSING"):
    send_notification({
        'method': 'task.update',
        'params': {
            'status': status
        }
    })


async def get_task():
    return await send_request({'method': 'task.get'})


def return_task():
    send_notification({
        'method': 'task.return'
    })
    sys.exit(0)


async def graphql(query, variables=None):
    return await send_request({
        'method': 'eywa.datasets.graphql',
        'params': {
            'query': query,
            'variables': variables
        }
    })


__stdin__task__ = None


def connect():
    __stdin__task__ = asyncio.create_task(read_stdin())


async def exit():

    if __stdin__task__ is not None:
        await __stdin__task__.cancel()
    sys.exit(0)


async def main():
    connect()
    await graphql("""
    searchTask (_limit:2000) {
      euuid
      status
      finished
      started
    }
    """)
    print(f'Exiting!')
    await exit()

asyncio.run(main())

# {"jsonrpc":"2.0","id":10,"result":{"searchTask":[{"euuid":"ba67467d-574e-4fdd-8f28-41e4c4937bc1","status":"EXCEPTION","data":{"input":"yum install epel-release -y"},"finished":null,"started":null},{"euuid":"878796bf-4cab-431c-a48f-c831348a49b8","status":"EXCEPTION","data":null,"finished":null,"started":null}]}}
